#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <functional>
using namespace std;

#define ever ;;

class DirectoryListener
{
	const char* directory;
	typedef vector<void(*)(const char*)> Listeners;
	Listeners addListeners;
	Listeners changeListeners;
	Listeners removeListeners;
	HANDLE directoryHandle;
	HANDLE thread;

	static DWORD WINAPI runListenLoop(LPVOID lpParam)
	{
		((DirectoryListener*)lpParam)->listenLoop();
		return 0;
	}

	bool listenLoop()
	{
		 directoryHandle = CreateFile(directory,FILE_LIST_DIRECTORY,
					   FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE,
					   NULL,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,NULL);

		if(directoryHandle==INVALID_HANDLE_VALUE)
		{
			printf("Could not read directory \"%s\"",directory);
			return false;
		}

		const DWORD maxBufferLength = 4096;
		char buffer[maxBufferLength];
		WCHAR wcFileName[MAX_PATH+1];
		char fileName[MAX_PATH+1];
		DWORD bufferLength = maxBufferLength;

		while(ReadDirectoryChangesW(directoryHandle,buffer,bufferLength,FALSE,FILE_NOTIFY_CHANGE_FILE_NAME,&bufferLength,NULL,NULL))
		{
			DWORD nextEntryOffset = 0;
			PFILE_NOTIFY_INFORMATION fileNotifyInformation = NULL;
			do
			{
				fileNotifyInformation = (PFILE_NOTIFY_INFORMATION)(buffer+nextEntryOffset);
				memcpy(wcFileName,fileNotifyInformation->FileName,fileNotifyInformation->FileNameLength);
				wcFileName[fileNotifyInformation->FileNameLength/sizeof(WCHAR)] = L'\0';
				wcstombs(fileName,wcFileName,MAX_PATH+1);
				Sleep(1000);
				Listeners* listeners = NULL;
				switch(fileNotifyInformation->Action)
				{
				case FILE_ACTION_ADDED:
					listeners = &addListeners;
					break;
				case FILE_ACTION_REMOVED:
					listeners = &removeListeners;
					break;
				case FILE_ACTION_MODIFIED:
					listeners = &changeListeners;
					break;
				}
				if(listeners)
				{
					for(Listeners::iterator iter(listeners->begin()); iter != listeners->end(); iter++)
					{
						(*iter)(fileName);
					}
				}
				nextEntryOffset += fileNotifyInformation->NextEntryOffset;
			}while(fileNotifyInformation->NextEntryOffset != 0);
		}
		CloseHandle(directoryHandle);
		return true;
	}
public:
	DirectoryListener(const char* directory)
		:directory(directory),
		thread(NULL),
		directoryHandle(NULL)
	{
	}
	void listen()
	{
		if(!thread)
		{
			thread = CreateThread(NULL,1024,&DirectoryListener::runListenLoop,this,NULL,NULL);
		}
	}
	
	void stop()
	{
		if(directoryHandle)
		{
			TerminateThread(thread,1);
			CloseHandle(directoryHandle);
			CloseHandle(thread);
			thread = NULL;
			directoryHandle = NULL;
		}
	}

	void addAddListener(void (*listenerFunction)(const char*))
	{
		addListeners.push_back(listenerFunction);
	}

	void addRemoveListener(void (*listenerFunction)(const char*))
	{
		removeListeners.push_back(listenerFunction);
	}

	void addChangeListener(void (*listenerFunction)(const char*))
	{
		changeListeners.push_back(listenerFunction);
	}
};